;
; Packet bridge -- Bridge over packet drivers
; (c) March 1994, Jon Beltran de Heredia AKA Yann/Iguana
;
; It should work on any packet driver above v1.07 with extended capabilities
;
; Change log:
; 22-03-94      Initial coding
; 23-03-94      Make the bridge TSR, & 'intelligent' with a bridging table
; 24-03-94      Fix it to detect when the bridging table gets full
;               I'll have to add some LRU algorithm!
; 25-03-94      Added a FIFO of packet buffers for each pkt drvr -
;               Now up to 2*40 packets can be buffered, so it's quite
;               difficult that the bridge will lose packets!
;

NUM_BT_ENTRIES  EQU 256
; As we have a 64K buffer for each pkt driver,  we have space for
; 65536/1600=40 packets
PKT_BUFFER_SIZE         EQU 1600
LAST_PKT_IN_BUFFER      EQU 40

CODE    SEGMENT PARA PUBLIC 'CODE'
        ASSUME CS:CODE, DS:CODE, SS:CODE, ES:CODE
        ORG     100h
start:  jmp     main

;
; Datos
;
        EVEN
overrun1        dw 0
overrun2        dw 0
sem_flush_b1    dw 0
sem_flush_b2    dw 0
b1_head         dw 0
b1_tail         dw 0
b2_head         dw 0
b2_tail         dw 0
buffer1_seg     dw ?
buffer2_seg     dw ?
bt_seg          dw ?    ; Bridging table start seg (offset=0)
bt_used_entries dw 0
pkt1_size       dw ?
pkt2_size       dw ?
;semaphore1      dw 0
semaphore2      dw 0
ip_type         db 08,00
arp_type        db 08,06
h_IP_60         dw ?
h_ARP_60        dw ?
h_IP_61         dw ?
h_ARP_61        dw ?

; Each item in the bridging tableis:
;     db 6 dup(?)       ; Ethernet physical address
;     db ?              ; 60h: it's in the network of the pktdrvr at 60h
;                       ; 61h: it's in the network of the pktdrvr at 61h
;     db ?              ; Not used


ENDLINE         EQU 0Dh, 0Ah
ENDSTR          EQU ENDLINE, '$'
CopyrightMsg    db ENDLINE
                db 'Packet Bridge   Version 0.2 beta   '
                db '(C) Marzo 1994', ENDLINE
                db 'Copyright  Jon Beltran de Heredia '
                db '- Grupo de seal y comunicaciones', ENDLINE
                db 'E.T.S.I.I. e I.T. de Bilbao - U.P.V./E.H.U.', ENDLINE
                db ENDLINE
                db 'Packet Bridge  0.2 beta bertsioa   '
                db '(C) 1994, Martxoan', ENDLINE
                db 'Copyright  Jon Beltran de Heredia '
                db '- Seinale eta komunikapen taldea', ENDLINE
                db 'Bilboko T.I. eta I.I.G.E.T. - E.H.U./U.P.V.', ENDLINE
                db ENDSTR
NoPktMsg        db 'Cannot find pkt drvrs in 60h and 61h!', ENDSTR
EscMsg          db 'Packet Bridge has been successfully installed', ENDSTR
Msg1            db 'Cannot realloc - not enough mem?', ENDSTR
Msg2            db 'Cannot alloc - I need about 144K bytes to run', ENDSTR
Msg3            db 'Cannot get info from 1st driver', ENDSTR
Msg4            db 'Cannot register IP on 1st driver', ENDSTR
Msg5            db 'Cannot register ARP on 1st driver', ENDSTR
Msg6            db 'Cannot get info from 2nd driver', ENDSTR
Msg7            db 'Cannot register IP on 2nd driver', ENDSTR
Msg8            db 'Cannot register ARP on 2nd driver', ENDSTR

;
; Macros
;
EXITMSG         MACRO   msg
                mov     dx,OFFSET msg
                mov     ah,9
                int     21h
                mov     ax,4CFFh
                int     21h
                ENDM

ASSERTNC        MACRO   msg
                LOCAL   no_error
                jnc     no_error
                EXITMSG msg
                no_error:
                ENDM

;==========================================================================
main:
; First of all, show (c) msg
        mov     dx,OFFSET CopyrightMsg
        mov     ah,9
        int     21h

; Setup memory map:
;
;       PSP & Code      4K
;       Stack           4K
;       Bridging table  8*NUM_BT_ENTRIES (=2K for 256 entries)
;       Buffer1         64K
;       Buffer2         64K

; Realloc to n paragraphs (where n=256+256+0.5*NUM_BT_ENTRIES+4096+4096)
        mov     bx,4096/16+4096/16+(NUM_BT_ENTRIES+1)/2+4096+4096
        mov     ah,4Ah
        int     21h
        ASSERTNC Msg1

  ; Set SP and SS for our new stack
        mov     ax,cs
        add     ax,256
        cli
        mov     ss,ax
        mov     sp,4096
        sti

  ; Set pointers to the bridging table, and to both packet buffers
        mov     ax,cs
        add     ax,256+256
        mov     [bt_seg],ax
        mov     ax,cs
        add     ax,256+256+(NUM_BT_ENTRIES+1)/2
        mov     [buffer1_seg],ax
        add     ax,4096
        mov     [buffer2_seg],ax

; Init DS
        mov     ax,cs
        mov     ds,ax

; Test if there are two packet drivers both in 60h and 61h
        xor     ax,ax
        mov     es,ax
        mov     bx,es:[60h*4]
        mov     es,es:[60h*4+2]
        add     bx,3
        cmp     WORD PTR es:[bx],'K'*256+'P'
        jnz     NoPktDrvrs
        xor     ax,ax
        mov     es,ax
        mov     bx,es:[61h*4]
        mov     es,es:[61h*4+2]
        add     bx,3
        cmp     WORD PTR es:[bx],'K'*256+'P'
        je      FoundPktDrvrs
NoPktDrvrs:
        mov     dx,OFFSET NoPktMsg
        mov     ah,9
        int     21h
        mov     ax,4CFFh
        int     21h
FoundPktDrvrs:

; Capture IP and ARP packets
        ; First get 60h pktdrv info
        mov     ah,1
        mov     al,255
        int     60h
        ASSERTNC Msg3
        mov     ax,cs
        mov     ds,ax
        mov     ah,2                    ; access_type
        mov     al,ch                   ; if. class
        mov     bx,dx                   ; if. type
        mov     dl,cl                   ; if. number
        mov     si,OFFSET ip_type
        push    cs
        pop     es                      ; receiver seg.
        mov     di,OFFSET rec_60        ; receiver off.
        push    ax
        push    bx
        push    dx
        push    es
        push    di
        mov     cx,2                    ; typelen
        int     60h
        ASSERTNC Msg4
        mov     [h_IP_60],ax            ; store handle
        ; ----- Set promiscuous mode for pktdrvr at 60h---------
        mov     ah,20
        mov     cx,6
        mov     bx,[h_IP_60]
        int     60h
        ; ------------------------------------------------------
        mov     ax,cs
        mov     ds,ax
        pop     di
        pop     es
        mov     si,OFFSET arp_type
        pop     dx
        pop     bx
        pop     ax
        mov     cx,2
        int     60h
        ASSERTNC Msg5
        mov     [h_ARP_60],ax           ; store handle
        ; Then get 61h pktdrv info
        mov     ah,1
        mov     al,255
        int     61h
        ASSERTNC Msg6
        mov     ax,cs
        mov     ds,ax
        mov     ah,2                    ; access_type
        mov     al,ch                   ; if. class
        mov     bx,dx                   ; if. type
        mov     dl,cl                   ; if. number
        mov     si,OFFSET ip_type
        push    cs
        pop     es                      ; receiver seg.
        mov     di,OFFSET rec_61        ; receiver off.
        push    ax
        push    bx
        push    dx
        push    es
        push    di
        mov     cx,2                    ; typelen
        int     61h
        ASSERTNC Msg7
        mov     [h_IP_61],ax            ; store handle
        ; ----- Set promiscuous mode for pktdrvr at 61h---------
        mov     ah,20
        mov     cx,6
        mov     bx,[h_IP_61]
        int     61h
        ; ------------------------------------------------------
        mov     ax,cs
        mov     ds,ax
        pop     di
        pop     es
        mov     si,OFFSET arp_type
        pop     dx
        pop     bx
        pop     ax
        mov     cx,2
        int     61h
        ASSERTNC Msg8
        mov     [h_ARP_61],ax            ; store handle

; Show initialization notification message
        mov     dx,OFFSET EscMsg
        mov     ah,9
        int     21h

; Now we have registered the ip and arp protocols, so the pkt drivers
; will call our functions when some pkt of those types arrives.
; We buffer them (40 buffers per pkt driver)

; Terminate and Stay Resident
        mov     ax,3100h        ; TSR service, ERRORLEVEL 0
        mov     dx,4096/16+4096/16+(NUM_BT_ENTRIES+1)/2+4096+4096
        int     21h             ; Return to DOS without unloading


COMMENT %
; We'll use this when we add uninstallation capability
; Unregister IP and ARP protocols from both interfaces
        mov     ah,3
        mov     bx,[h_IP_60]
        int     60h
        mov     ah,3
        mov     bx,[h_ARP_60]
        int     60h
        mov     ah,3
        mov     bx,[h_IP_61]
        int     61h
        mov     ah,3
        mov     bx,[h_ARP_61]
        int     61h

; Return to DOS
        mov     ax,4C00h
        int     21h
%

        ASSUME cs:code, ds:nothing, es:nothing, ss:nothing
;
; ISR routine for IP and ARP packets from pkt drv at 60h
;
rec_60:
        cld
        push    ax
        push    cx
        push    ds
        push    si
        call    FlushBuffer1            ; Try to flush the buffer
        pop     si
        pop     ds
        pop     cx
        pop     ax
        or      ax,ax
        jz      WantsBuffer1
        ; So we received a packet!
        ; First -> add source address to bridging table if necessary
        ; DS:SI points to the buffer
        push    si
        add     si,6
        call    LookupAddress
        jnc     KnownAddr1              ; If already in table, don't add
        mov     al,60h
        call    AddAddress              ; LookupAddress doesn't change
KnownAddr1:                             ;  neither DS nor SI, so don't reload
        pop     si
        push    si
        ; Then -> send pkt to 61h if necessary
        call    LookupAddress
        jc      NeedaSend1              ; If address not known, bridge it!
        cmp     al,60h
        jne     NeedaSend1
        ; Don't send 'cos it's in the 60h network
        pop     si
        mov     WORD PTR [si-2],0       ; Mark as not necessary to send
        iret

NeedaSend1:
        pop     si
        xor     BYTE PTR [si-1],80h     ; Mark as received!
        call    FlushBuffer1
        iret

WantsBuffer1:
        ; Return the Buffer1 address if there is space available...
        call    AllocBuffer1
        iret

;
; ISR routine for IP and ARP packets from pkt drv at 61h
;
rec_61:
        cld
        push    ax
        push    cx
        push    ds
        push    si
        call    FlushBuffer2            ; Try to flush the buffer
        pop     si
        pop     ds
        pop     cx
        pop     ax
        or      ax,ax
        jz      WantsBuffer2
        ; So we received a packet!
        ; First -> add source address to bridging table if necessary
        ; DS:SI points to the buffer
        push    si
        add     si,6
        call    LookupAddress
        jnc     KnownAddr2              ; If already in table, don't add
        mov     al,61h
        call    AddAddress              ; LookupAddress doesn't change
KnownAddr2:                             ;  neither DS nor SI, so don't reload
        pop     si
        push    si
        ; Then -> send pkt to 60h if necessary
        call    LookupAddress
        jc      NeedaSend2              ; If address not known, bridge it!
        cmp     al,61h
        jne     NeedaSend2
        ; Don't send 'cos it's in the 61h network
        pop     si
        mov     WORD PTR [si-2],0       ; Mark as not necessary to send
        iret

NeedaSend2:
        pop     si
        xor     BYTE PTR [si-1],80h     ; Mark as received!
        call    FlushBuffer2
        iret

WantsBuffer2:
        ; Return the Buffer2 address if there is space available...
        call    AllocBuffer2
        iret

;
; Function to lookup an address in the bridging table
; In:   DS:SI -> 6 byte Ethernet address
; Out:  Carry:   Not found
;       NCarry:  Found
;           AL:  60h or 61h, the int of the pktdrvr towards the requested node
;
LookupAddress:
        mov     ax,[bt_seg]
        mov     es,ax
        xor     di,di
        mov     bp,[bt_used_entries]
        or      bp,bp
        jz      LANotFound
LALoop: push    si
        push    di
        mov     cx,3
        repe    cmpsw
        je      LAFound
        pop     di
        pop     si
        add     di,8
        dec     bp
        jnz     LALoop
LANotFound:
        stc
        ret
LAFound:pop     di
        pop     si
        mov     al,es:[di+6]
        clc
        ret

;
; Function to add an address to the bridging table
; In:   DS:SI -> 6 byte Ethernet address
;          AL -> 60h or 61h, int of the pktdrvr to store in the table
; Out:  void
;
AddAddress:
        cmp     [bt_used_entries],NUM_BT_ENTRIES
        je      DontAdd
        mov     bp,si
        mov     dl,al
        mov     es,[bt_seg]
        mov     bx,[bt_used_entries]
        add     bx,bx
        add     bx,bx
        add     bx,bx
        ; Now es:bx -> New entry
        lodsw
        mov     es:[bx],ax
        lodsw
        mov     es:[bx+2],ax
        lodsw
        mov     es:[bx+4],ax
        mov     es:[bx+6],dl
        inc     [bt_used_entries]
        mov     si,bp
DontAdd:
        ret

;
; Function to flush the packets stored in buffer 1
;
FlushBuffer1:
        cmp     [sem_flush_b1],0
        je      NotAlreadyFlushing1
        ret
NotAlreadyFlushing1:
        mov     [sem_flush_b1],1
        mov     bx,[b1_tail]
        cmp     bx,[b1_head]
        jne     NotBuf1Empty
        mov     [sem_flush_b1],0
        ret
NotBuf1Empty:
        mov     ds,[buffer1_seg]
        mov     bx,[b1_tail]
        mov     ax,PKT_BUFFER_SIZE
        mul     bx
        mov     si,ax                   ; ds:si -> first pkt to flush
                                        ;  word len, pkt data...
FlushLoop1:
        push    si
        push    ds
        mov     cx,[si]
        or      cx,cx
        jns     NotNeg1                 ; Packet not ready to send
        add     sp,4
        jmp     EndFlushing1
NotNeg1:
        jz      SkipPkt1
        inc     si
        inc     si
        mov     ah,4                    ; Send pkt f(), ds:si->pkt, cx:len
        int     61h
SkipPkt1:
        pop     ds
        pop     si
        add     si,PKT_BUFFER_SIZE
        mov     ax,[b1_tail]
        inc     ax
        cmp     ax,LAST_PKT_IN_BUFFER
        je      RingWrap1
TestEndFlush1:
        mov     [b1_tail],ax
        mov     [overrun1],0            ; Just sent pkt -> no overrun possible
        cmp     ax,[b1_head]
        jne     FlushLoop1
        ; So we already flushed all the buffer1...
EndFlushing1:
        mov     [sem_flush_b1],0
        ret

RingWrap1:
        xor     si,si                   ; The whole buffer starts at ds:0
        xor     ax,ax
        jmp     TestEndFlush1

;
; Function to allocate the next free pkt buffer in buffer1
;
; In:   CX: length (always less than 1598 bytes, there should be no
;                   problem with Ethernet)
; Out:  ES:DI -> Buffer for the packet
;
AllocBuffer1:
        cmp     [overrun1],0
        jne     b1_overrun
        mov     bx,[b1_head]
        mov     ax,bx
        inc     ax
        cmp     ax,LAST_PKT_IN_BUFFER
        jne     AllocDontWrap1
        xor     ax,ax
AllocDontWrap1:
        cmp     ax,[b1_tail]            ; Are we going to overrun the ring?
        jne     NoOverrun1
        mov     [overrun1],1
        ; Next pkts before flushing will be dropped!
NoOverrun1:
        mov     [b1_head],ax
        mov     es,[buffer1_seg]
        mov     ax,PKT_BUFFER_SIZE
        mul     bx
        mov     di,ax
        xor     ch,80h
        mov     es:[di],cx      ; Negative length to mark it's not
                                ;  ready to send yet
        inc     di
        inc     di
        ret

; We have to drop the packet, there are 40 of them unflushed!
b1_Overrun:
        xor     ax,ax
        mov     es,ax
        mov     di,ax
        ret

;
; Function to flush the packets stored in buffer 2
;
FlushBuffer2:
        cmp     [sem_flush_b2],0
        je      NotAlreadyFlushing2
        ret
NotAlreadyFlushing2:
        mov     [sem_flush_b2],1
        mov     bx,[b2_tail]
        cmp     bx,[b2_head]
        jne     NotBuf2Empty
        mov     [sem_flush_b2],0
        ret
NotBuf2Empty:
        mov     ds,[buffer2_seg]
        mov     bx,[b2_tail]
        mov     ax,PKT_BUFFER_SIZE
        mul     bx
        mov     si,ax                   ; ds:si -> first pkt to flush
                                        ;  word len, pkt data...
FlushLoop2:
        push    si
        push    ds
        mov     cx,[si]
        or      cx,cx
        jns     NotNeg2                 ; Packet not ready to send
        add     sp,4
        jmp     EndFlushing2
NotNeg2:
        jz      SkipPkt2
        inc     si
        inc     si
        mov     ah,4                    ; Send pkt f(), ds:si->pkt, cx:len
        int     60h
SkipPkt2:
        pop     ds
        pop     si
        add     si,PKT_BUFFER_SIZE
        mov     ax,[b2_tail]
        inc     ax
        cmp     ax,LAST_PKT_IN_BUFFER
        je      RingWrap2
TestEndFlush2:
        mov     [b2_tail],ax
        mov     [overrun2],0            ; Just sent pkt -> no overrun possible
        cmp     ax,[b2_head]
        jne     FlushLoop2
        ; So we already flushed all the buffer2...
EndFlushing2:
        mov     [sem_flush_b2],0
        ret

RingWrap2:
        xor     si,si                   ; The whole buffer starts at ds:0
        xor     ax,ax
        jmp     TestEndFlush2

;
; Function to allocate the next free pkt buffer in buffer2
;
; In:   CX: length (always less than 1598 bytes, there should be no
;                   problem with Ethernet)
; Out:  ES:DI -> Buffer for the packet
;
AllocBuffer2:
        cmp     [overrun2],0
        jne     b2_overrun
        mov     bx,[b2_head]
        mov     ax,bx
        inc     ax
        cmp     ax,LAST_PKT_IN_BUFFER
        jne     AllocDontWrap2
        xor     ax,ax
AllocDontWrap2:
        cmp     ax,[b2_tail]            ; Are we going to overrun the ring?
        jne     NoOverrun2
        mov     [overrun2],1
        ; Next pkts before flushing will be dropped!
NoOverrun2:
        mov     [b2_head],ax
        mov     es,[buffer2_seg]
        mov     ax,PKT_BUFFER_SIZE
        mul     bx
        mov     di,ax
        xor     ch,80h
        mov     es:[di],cx      ; Negative length to mark it's not
                                ;  ready to send yet
        inc     di
        inc     di
        ret

; We have to drop the packet, there are 40 of them unflushed!
b2_Overrun:
        xor     ax,ax
        mov     es,ax
        mov     di,ax
        ret


CODE    ENDS

        END     start
