;
; Packet bridge -- Bridge over packet drivers
; (c) March 1994, Jon Beltran de Heredia AKA Yann/Iguana
;
; It should work on any packet driver above v1.07 with extended capabilities
;
; Change log:
; 22-03-94      Initial coding
; 23-03-94      Make the bridge TSR, & 'intelligent' with a bridging table
; 24-03-94      Fix it to detect when the bridging table gets full
;               I'll have to add some LRU algorithm!
; 25-03-94      Added a FIFO of packet buffers for each pkt drvr -
;               Now up to 2*40 packets can be buffered, so it's quite
;               difficult that the bridge will lose packets!
; 26-01-2000	Same as pktbri.asm, but converted to NASM format
;		for greater portability.
;

NUM_BT_ENTRIES  EQU 256
; As we have a 64K buffer for each pkt driver,  we have space for
; 65536/1600=40 packets
PKT_BUFFER_SIZE         EQU 1600
LAST_PKT_IN_BUFFER      EQU 40

	section	.text
        org     0x100
start:  jmp     main

;
; Datos
;
        align 2
overrun1        dw 0
overrun2        dw 0
sem_flush_b1    dw 0
sem_flush_b2    dw 0
b1_head         dw 0
b1_tail         dw 0
b2_head         dw 0
b2_tail         dw 0
buffer1_seg     dw 0
buffer2_seg     dw 0
bt_seg          dw 0    ; Bridging table start seg (offset=0)
bt_used_entries dw 0
pkt1_size       dw 0
pkt2_size       dw 0
;semaphore1      dw 0
semaphore2      dw 0
ip_type         db 08,00
arp_type        db 08,06
h_IP_60         dw 0
h_ARP_60        dw 0
h_IP_61         dw 0
h_ARP_61        dw 0

; Each item in the bridging tableis:
;     db 6 dup(?)       ; Ethernet physical address
;     db ?              ; 0x60: it's in the network of the pktdrvr at 0x60
;                       ; 0x61: it's in the network of the pktdrvr at 0x61
;     db ?              ; Not used


%define	ENDLINE 0x0D, 0x0A
%define	ENDSTR  ENDLINE, '$'

CopyrightMsg    db ENDLINE
                db 'Packet Bridge   Version 0.2 beta   '
                db '(C) Marzo 1994', ENDLINE
                db 'Copyright  Jon Beltran de Heredia '
                db '- Grupo de seal y comunicaciones', ENDLINE
                db 'E.T.S.I.I. e I.T. de Bilbao - U.P.V./E.H.U.', ENDLINE
                db ENDLINE
                db 'Packet Bridge  0.2 beta bertsioa   '
                db '(C) 1994, Martxoan', ENDLINE
                db 'Copyright  Jon Beltran de Heredia '
                db '- Seinale eta komunikapen taldea', ENDLINE
                db 'Bilboko T.I. eta I.I.G.E.T. - E.H.U./U.P.V.', ENDLINE
                db ENDSTR
NoPktMsg        db 'Cannot find pkt drvrs in 60h and 61h!', ENDSTR
EscMsg          db 'Packet Bridge has been successfully installed', ENDSTR
Msg1            db 'Cannot realloc - not enough mem?', ENDSTR
Msg2            db 'Cannot alloc - I need about 144K bytes to run', ENDSTR
Msg3            db 'Cannot get info from 1st driver', ENDSTR
Msg4            db 'Cannot register IP on 1st driver', ENDSTR
Msg5            db 'Cannot register ARP on 1st driver', ENDSTR
Msg6            db 'Cannot get info from 2nd driver', ENDSTR
Msg7            db 'Cannot register IP on 2nd driver', ENDSTR
Msg8            db 'Cannot register ARP on 2nd driver', ENDSTR

;
; Macros
;
%macro	EXITMSG 1
                mov     dx,%1
                mov     ah,9
                int     0x21
                mov     ax,0x4CFF
                int     0x21
%endmacro

%macro	ASSERTNC 1
                jnc     %%no_error
                EXITMSG %1
                %%no_error:
%endmacro

;==========================================================================
main:
; First of all, show (c) msg
        mov     dx,CopyrightMsg
        mov     ah,9
        int     0x21

; Setup memory map:
;
;       PSP & Code      4K
;       Stack           4K
;       Bridging table  8*NUM_BT_ENTRIES (=2K for 256 entries)
;       Buffer1         64K
;       Buffer2         64K

; Realloc to n paragraphs (where n=256+256+0.5*NUM_BT_ENTRIES+4096+4096)
        mov     bx,4096/16+4096/16+(NUM_BT_ENTRIES+1)/2+4096+4096
        mov     ah,0x4A
        int     0x21
        ASSERTNC Msg1

  ; Set SP and SS for our new stack
        mov     ax,cs
        add     ax,256
        cli
        mov     ss,ax
        mov     sp,4096
        sti

  ; Set pointers to the bridging table, and to both packet buffers
        mov     ax,cs
        add     ax,256+256
        mov     [bt_seg],ax
        mov     ax,cs
        add     ax,256+256+(NUM_BT_ENTRIES+1)/2
        mov     [buffer1_seg],ax
        add     ax,4096
        mov     [buffer2_seg],ax

; Init DS
        mov     ax,cs
        mov     ds,ax

; Test if there are two packet drivers both in 0x60 and 0x61
        xor     ax,ax
        mov     es,ax
        mov     bx,[es:0x60*4]
        mov     es,[es:0x60*4+2]
        add     bx,BYTE 3
        cmp     WORD [es:bx],'K'*256+'P'
        jnz     NoPktDrvrs
        xor     ax,ax
        mov     es,ax
        mov     bx,[es:0x61*4]
        mov     es,[es:0x61*4+2]
        add     bx,BYTE 3
        cmp     WORD [es:bx],'K'*256+'P'
        je      FoundPktDrvrs
NoPktDrvrs:
        mov     dx,NoPktMsg
        mov     ah,9
        int     0x21
        mov     ax,4CFFh
        int     0x21
FoundPktDrvrs:

; Capture IP and ARP packets
        ; First get 0x60 pktdrv info
        mov     ah,1
        mov     al,255
        int     0x60
        ASSERTNC Msg3
        mov     ax,cs
        mov     ds,ax
        mov     ah,2                    ; access_type
        mov     al,ch                   ; if. class
        mov     bx,dx                   ; if. type
        mov     dl,cl                   ; if. number
        mov     si,ip_type
        push    cs
        pop     es                      ; receiver seg.
        mov     di,rec_60        ; receiver off.
        push    ax
        push    bx
        push    dx
        push    es
        push    di
        mov     cx,2                    ; typelen
        int     0x60
        ASSERTNC Msg4
        mov     [h_IP_60],ax            ; store handle
        ; ----- Set promiscuous mode for pktdrvr at 0x60---------
        mov     ah,20
        mov     cx,6
        mov     bx,[h_IP_60]
        int     0x60
        ; ------------------------------------------------------
        mov     ax,cs
        mov     ds,ax
        pop     di
        pop     es
        mov     si,arp_type
        pop     dx
        pop     bx
        pop     ax
        mov     cx,2
        int     0x60
        ASSERTNC Msg5
        mov     [h_ARP_60],ax           ; store handle
        ; Then get 0x61 pktdrv info
        mov     ah,1
        mov     al,255
        int     0x61
        ASSERTNC Msg6
        mov     ax,cs
        mov     ds,ax
        mov     ah,2                    ; access_type
        mov     al,ch                   ; if. class
        mov     bx,dx                   ; if. type
        mov     dl,cl                   ; if. number
        mov     si,ip_type
        push    cs
        pop     es                      ; receiver seg.
        mov     di,rec_61        ; receiver off.
        push    ax
        push    bx
        push    dx
        push    es
        push    di
        mov     cx,2                    ; typelen
        int     0x61
        ASSERTNC Msg7
        mov     [h_IP_61],ax            ; store handle
        ; ----- Set promiscuous mode for pktdrvr at 0x61---------
        mov     ah,20
        mov     cx,6
        mov     bx,[h_IP_61]
        int     0x61
        ; ------------------------------------------------------
        mov     ax,cs
        mov     ds,ax
        pop     di
        pop     es
        mov     si,arp_type
        pop     dx
        pop     bx
        pop     ax
        mov     cx,2
        int     0x61
        ASSERTNC Msg8
        mov     [h_ARP_61],ax            ; store handle

; Show initialization notification message
        mov     dx,EscMsg
        mov     ah,9
        int     0x21

; Now we have registered the ip and arp protocols, so the pkt drivers
; will call our functions when some pkt of those types arrives.
; We buffer them (40 buffers per pkt driver)

; Terminate and Stay Resident
        mov     ax,3100h        ; TSR service, ERRORLEVEL 0
        mov     dx,4096/16+4096/16+(NUM_BT_ENTRIES+1)/2+4096+4096
        int     0x21             ; Return to DOS without unloading


%if	0
; We'll use this when we add uninstallation capability
; Unregister IP and ARP protocols from both interfaces
        mov     ah,3
        mov     bx,[h_IP_60]
        int     0x60
        mov     ah,3
        mov     bx,[h_ARP_60]
        int     0x60
        mov     ah,3
        mov     bx,[h_IP_61]
        int     0x61
        mov     ah,3
        mov     bx,[h_ARP_61]
        int     0x61

; Return to DOS
        mov     ax,4C00h
        int     0x21
%endif

; These interrupt routines require addressing the data from cs.
;
; ISR routine for IP and ARP packets from pkt drv at 0x60
;
rec_60:
        cld
        push    ax
        push    cx
        push    ds
        push    si
        call    FlushBuffer1            ; Try to flush the buffer
        pop     si
        pop     ds
        pop     cx
        pop     ax
        or      ax,ax
        jz      WantsBuffer1
        ; So we received a packet!
        ; First -> add source address to bridging table if necessary
        ; DS:SI points to the buffer
        push    si
        add     si,BYTE 6
        call    LookupAddress
        jnc     KnownAddr1              ; If already in table, don't add
        mov     al,0x60
        call    AddAddress              ; LookupAddress doesn't change
KnownAddr1:                             ;  neither DS nor SI, so don't reload
        pop     si
        push    si
        ; Then -> send pkt to 0x61 if necessary
        call    LookupAddress
        jc      NeedaSend1              ; If address not known, bridge it!
        cmp     al,0x60
        jne     NeedaSend1
        ; Don't send 'cos it's in the 0x60 network
        pop     si
        mov     WORD [si-2],0       ; Mark as not necessary to send
        iret

NeedaSend1:
        pop     si
        xor     BYTE [si-1],0x80     ; Mark as received!
        call    FlushBuffer1
        iret

WantsBuffer1:
        ; Return the Buffer1 address if there is space available...
        call    AllocBuffer1
        iret

;
; ISR routine for IP and ARP packets from pkt drv at 0x61
;
rec_61:
        cld
        push    ax
        push    cx
        push    ds
        push    si
        call    FlushBuffer2            ; Try to flush the buffer
        pop     si
        pop     ds
        pop     cx
        pop     ax
        or      ax,ax
        jz      WantsBuffer2
        ; So we received a packet!
        ; First -> add source address to bridging table if necessary
        ; DS:SI points to the buffer
        push    si
        add     si,BYTE 6
        call    LookupAddress
        jnc     KnownAddr2              ; If already in table, don't add
        mov     al,0x61
        call    AddAddress              ; LookupAddress doesn't change
KnownAddr2:                             ;  neither DS nor SI, so don't reload
        pop     si
        push    si
        ; Then -> send pkt to 0x60 if necessary
        call    LookupAddress
        jc      NeedaSend2              ; If address not known, bridge it!
        cmp     al,0x61
        jne     NeedaSend2
        ; Don't send 'cos it's in the 0x61 network
        pop     si
        mov     WORD [si-2],0       ; Mark as not necessary to send
        iret

NeedaSend2:
        pop     si
        xor     BYTE [si-1],0x80     ; Mark as received!
        call    FlushBuffer2
        iret

WantsBuffer2:
        ; Return the Buffer2 address if there is space available...
        call    AllocBuffer2
        iret

;
; Function to lookup an address in the bridging table
; In:   DS:SI -> 6 byte Ethernet address
; Out:  Carry:   Not found
;       NCarry:  Found
;           AL:  0x60 or 0x61, the int of the pktdrvr towards the requested node
;
LookupAddress:
        mov     ax,[cs:bt_seg]
        mov     es,ax
        xor     di,di
        mov     bp,[cs:bt_used_entries]
        or      bp,bp
        jz      LANotFound
LALoop: push    si
        push    di
        mov     cx,3
        repe    cmpsw
        je      LAFound
        pop     di
        pop     si
        add     di,BYTE 8
        dec     bp
        jnz     LALoop
LANotFound:
        stc
        ret
LAFound:pop     di
        pop     si
        mov     al,[es:di+6]
        clc
        ret

;
; Function to add an address to the bridging table
; In:   DS:SI -> 6 byte Ethernet address
;          AL -> 0x60 or 0x61, int of the pktdrvr to store in the table
; Out:  void
;
AddAddress:
        cmp     [cs:bt_used_entries],WORD NUM_BT_ENTRIES
        je      DontAdd
        mov     bp,si
        mov     dl,al
        mov     es,[cs:bt_seg]
        mov     bx,[cs:bt_used_entries]
        add     bx,bx
        add     bx,bx
        add     bx,bx
        ; Now es:bx -> New entry
        lodsw
        mov     [es:bx],ax
        lodsw
        mov     [es:bx+2],ax
        lodsw
        mov     [es:bx+4],ax
        mov     [es:bx+6],dl
        inc     WORD [cs:bt_used_entries]
        mov     si,bp
DontAdd:
        ret

;
; Function to flush the packets stored in buffer 1
;
FlushBuffer1:
        cmp     WORD [cs:sem_flush_b1],BYTE 0
        je      NotAlreadyFlushing1
        ret
NotAlreadyFlushing1:
        mov     [cs:sem_flush_b1],WORD 1
        mov     bx,[cs:b1_tail]
        cmp     bx,[cs:b1_head]
        jne     NotBuf1Empty
        mov     [cs:sem_flush_b1],WORD 0
        ret
NotBuf1Empty:
        mov     ds,[cs:buffer1_seg]
        mov     bx,[cs:b1_tail]
        mov     ax,PKT_BUFFER_SIZE
        mul     bx
        mov     si,ax                   ; ds:si -> first pkt to flush
                                        ;  word len, pkt data...
FlushLoop1:
        push    si
        push    ds
        mov     cx,[si]
        or      cx,cx
        jns     NotNeg1                 ; Packet not ready to send
        add     sp,BYTE 4
        jmp     EndFlushing1
NotNeg1:
        jz      SkipPkt1
        inc     si
        inc     si
        mov     ah,4                    ; Send pkt f(), ds:si->pkt, cx:len
        int     0x61
SkipPkt1:
        pop     ds
        pop     si
        add     si,PKT_BUFFER_SIZE
        mov     ax,[cs:b1_tail]
        inc     ax
        cmp     ax,LAST_PKT_IN_BUFFER
        je      RingWrap1
TestEndFlush1:
        mov     [cs:b1_tail],ax
        mov     [cs:overrun1],WORD 0            ; Just sent pkt -> no overrun possible
        cmp     ax,[cs:b1_head]
        jne     FlushLoop1
        ; So we already flushed all the buffer1...
EndFlushing1:
        mov     [cs:sem_flush_b1],WORD 0
        ret

RingWrap1:
        xor     si,si                   ; The whole buffer starts at ds:0
        xor     ax,ax
        jmp     SHORT TestEndFlush1

;
; Function to allocate the next free pkt buffer in buffer1
;
; In:   CX: length (always less than 1598 bytes, there should be no
;                   problem with Ethernet)
; Out:  ES:DI -> Buffer for the packet
;
AllocBuffer1:
        cmp     WORD [cs:overrun1],BYTE 0
        jne     b1_Overrun
        mov     bx,[cs:b1_head]
        mov     ax,bx
        inc     ax
        cmp     ax,LAST_PKT_IN_BUFFER
        jne     AllocDontWrap1
        xor     ax,ax
AllocDontWrap1:
        cmp     ax,[cs:b1_tail]            ; Are we going to overrun the ring?
        jne     NoOverrun1
        mov     [cs:overrun1],WORD 1
        ; Next pkts before flushing will be dropped!
NoOverrun1:
        mov     [cs:b1_head],ax
        mov     es,[cs:buffer1_seg]
        mov     ax,PKT_BUFFER_SIZE
        mul     bx
        mov     di,ax
        xor     ch,0x80
        mov     [es:di],cx      ; Negative length to mark it's not
                                ;  ready to send yet
        inc     di
        inc     di
        ret

; We have to drop the packet, there are 40 of them unflushed!
b1_Overrun:
        xor     ax,ax
        mov     es,ax
        mov     di,ax
        ret

;
; Function to flush the packets stored in buffer 2
;
FlushBuffer2:
        cmp     WORD [cs:sem_flush_b2],BYTE 0
        je      NotAlreadyFlushing2
        ret
NotAlreadyFlushing2:
        mov     [cs:sem_flush_b2],WORD 1
        mov     bx,[cs:b2_tail]
        cmp     bx,[cs:b2_head]
        jne     NotBuf2Empty
        mov     [cs:sem_flush_b2],WORD 0
        ret
NotBuf2Empty:
        mov     ds,[cs:buffer2_seg]
        mov     bx,[cs:b2_tail]
        mov     ax,PKT_BUFFER_SIZE
        mul     bx
        mov     si,ax                   ; ds:si -> first pkt to flush
                                        ;  word len, pkt data...
FlushLoop2:
        push    si
        push    ds
        mov     cx,[si]
        or      cx,cx
        jns     NotNeg2                 ; Packet not ready to send
        add     sp,BYTE 4
        jmp     EndFlushing2
NotNeg2:
        jz      SkipPkt2
        inc     si
        inc     si
        mov     ah,4                    ; Send pkt f(), ds:si->pkt, cx:len
        int     0x60
SkipPkt2:
        pop     ds
        pop     si
        add     si,PKT_BUFFER_SIZE
        mov     ax,[cs:b2_tail]
        inc     ax
        cmp     ax,LAST_PKT_IN_BUFFER
        je      RingWrap2
TestEndFlush2:
        mov     [cs:b2_tail],ax
        mov     [cs:overrun2],WORD 0            ; Just sent pkt -> no overrun possible
        cmp     ax,[cs:b2_head]
        jne     FlushLoop2
        ; So we already flushed all the buffer2...
EndFlushing2:
        mov     [cs:sem_flush_b2],WORD 0
        ret

RingWrap2:
        xor     si,si                   ; The whole buffer starts at ds:0
        xor     ax,ax
        jmp     SHORT TestEndFlush2

;
; Function to allocate the next free pkt buffer in buffer2
;
; In:   CX: length (always less than 1598 bytes, there should be no
;                   problem with Ethernet)
; Out:  ES:DI -> Buffer for the packet
;
AllocBuffer2:
        cmp     WORD [cs:overrun2],BYTE 0
        jne     b2_Overrun
        mov     bx,[cs:b2_head]
        mov     ax,bx
        inc     ax
        cmp     ax,LAST_PKT_IN_BUFFER
        jne     AllocDontWrap2
        xor     ax,ax
AllocDontWrap2:
        cmp     ax,[cs:b2_tail]            ; Are we going to overrun the ring?
        jne     NoOverrun2
        mov     [cs:overrun2],WORD 1
        ; Next pkts before flushing will be dropped!
NoOverrun2:
        mov     [cs:b2_head],ax
        mov     es,[cs:buffer2_seg]
        mov     ax,PKT_BUFFER_SIZE
        mul     bx
        mov     di,ax
        xor     ch,0x80
        mov     [es:di],cx      ; Negative length to mark it's not
                                ;  ready to send yet
        inc     di
        inc     di
        ret

; We have to drop the packet, there are 40 of them unflushed!
b2_Overrun:
        xor     ax,ax
        mov     es,ax
        mov     di,ax
        ret
